@@ -8,25 +8,26 @@ import (
 	"io/fs"
 	"os"
 	"path/filepath"
+	"slices"
 	"strings"
 	"sync"
 
-	"github.com/gruntwork-io/terragrunt/config/hclparse"
 	"github.com/gruntwork-io/terragrunt/internal/worktrees"
+	"github.com/gruntwork-io/terragrunt/pkg/config/hclparse"
 
 	"github.com/gruntwork-io/terragrunt/internal/component"
 	"github.com/gruntwork-io/terragrunt/internal/experiment"
 	"github.com/gruntwork-io/terragrunt/internal/filter"
 	"github.com/gruntwork-io/terragrunt/internal/report"
-	"github.com/gruntwork-io/terragrunt/shell"
-	"github.com/gruntwork-io/terragrunt/util"
+	"github.com/gruntwork-io/terragrunt/internal/shell"
+	"github.com/gruntwork-io/terragrunt/internal/util"
 
-	"github.com/gruntwork-io/terragrunt/telemetry"
+	"github.com/gruntwork-io/terragrunt/internal/telemetry"
 
-	"github.com/gruntwork-io/terragrunt/config"
 	"github.com/gruntwork-io/terragrunt/internal/errors"
-	"github.com/gruntwork-io/terragrunt/options"
+	"github.com/gruntwork-io/terragrunt/pkg/config"
 	"github.com/gruntwork-io/terragrunt/pkg/log"
+	"github.com/gruntwork-io/terragrunt/pkg/options"
 	"github.com/hashicorp/hcl/v2"
 	"github.com/zclconf/go-cty/cty"
 	"golang.org/x/sync/errgroup"
@@ -249,7 +250,7 @@ func (d *Discovery) SetParseOptions(options []hclparse.Option) {
 // WithOptions ingests runner options and applies any discovery-relevant settings.
 // Currently, it extracts HCL parser options provided via common.ParseOptionsProvider
 // and forwards them to discovery's parser configuration.
-func (d *Discovery) WithOptions(opts ...interface{}) *Discovery {
+func (d *Discovery) WithOptions(opts ...any) *Discovery {
 	var parserOptions []hclparse.Option
 
 	for _, opt := range opts {
@@ -415,6 +416,7 @@ func Parse(
 		config.FeatureFlagsBlock,
 		config.ExcludeBlock,
 		config.ErrorsBlock,
+		config.TerragruntVersionConstraints,
 	).WithSkipOutputsResolution()
 
 	// Apply custom parser options if provided via discovery
@@ -424,16 +426,16 @@ func Parse(
 
 	if suppressParseErrors {
 		// Suppressing parse errors to avoid false positive errors
-		parseOptions := append(
+		parseOptions := slices.Concat(
 			parsingCtx.ParserOptions,
-			hclparse.WithDiagnosticsHandler(func(
+			[]hclparse.Option{hclparse.WithDiagnosticsHandler(func(
 				file *hcl.File,
 				hclDiags hcl.Diagnostics,
 			) (hcl.Diagnostics, error) {
 				l.Debugf("Suppressed parsing errors %w", hclDiags)
 
 				return nil, nil
-			}))
+			})})
 		parsingCtx = parsingCtx.WithParseOption(parseOptions)
 	}
 
@@ -488,12 +490,33 @@ func Parse(
 	// Populate the Reading field with files read during parsing.
 	// The parsing context tracks all files that were read.
 	if parsingCtx.FilesRead != nil {
-		c.SetReading(*parsingCtx.FilesRead...)
+		readFiles := sanitizeReadFiles(*parsingCtx.FilesRead)
+
+		c.SetReading(readFiles...)
 	}
 
 	return nil
 }
 
+// sanitizeReadFiles sanitizes the list of files being read by a component.
+//
+// It removes empty files, sorts and deduplicates the list.
+func sanitizeReadFiles(files []string) []string {
+	if len(files) == 0 {
+		return []string{}
+	}
+
+	files = slices.Clone(files)
+
+	files = slices.DeleteFunc(files, func(file string) bool {
+		return len(file) == 0
+	})
+
+	slices.Sort(files)
+
+	return slices.Compact(files)
+}
+
 // isInHiddenDirectory returns true if the path is in a hidden directory.
 func (d *Discovery) isInHiddenDirectory(hiddenDirMemo *hiddenDirMemo, path string) bool {
 	ok := hiddenDirMemo.contains(path)
@@ -753,7 +776,10 @@ func (d *Discovery) createComponentFromPath(path string, filenames []string) com
 		}
 
 		if d.discoveryContext != nil {
-			c.SetDiscoveryContext(d.discoveryContext)
+			discoveryCtx := d.discoveryContext.Copy()
+			discoveryCtx.SuggestOrigin(component.OriginPathDiscovery)
+
+			c.SetDiscoveryContext(discoveryCtx)
 		}
 
 		return c
@@ -958,10 +984,6 @@ func (d *Discovery) Discover(
 						WithMaxDepth(d.maxDependencyDepth).
 						WithNumWorkers(d.numWorkers)
 
-					if d.discoveryContext != nil {
-						dependencyDiscovery = dependencyDiscovery.WithDiscoveryContext(d.discoveryContext)
-					}
-
 					if d.suppressParseErrors {
 						dependencyDiscovery = dependencyDiscovery.WithSuppressParseErrors()
 					}
@@ -1004,10 +1026,6 @@ func (d *Discovery) Discover(
 						WithMaxDepth(d.maxDependencyDepth).
 						WithNumWorkers(d.numWorkers)
 
-					if d.discoveryContext != nil {
-						dependentDiscovery = dependentDiscovery.WithDiscoveryContext(d.discoveryContext)
-					}
-
 					if d.suppressParseErrors {
 						dependentDiscovery = dependentDiscovery.WithSuppressParseErrors()
 					}
@@ -1211,9 +1229,7 @@ func buildDependentsIndex(components component.Components) map[string][]string {
 
 		for _, dep := range c.Dependencies() {
 			depPath := resolvePath(dep.Path())
-			dependentUnits[depPath] = util.RemoveDuplicatesFromList(
-				append(dependentUnits[depPath], cPath),
-			)
+			dependentUnits[depPath] = util.RemoveDuplicates(append(dependentUnits[depPath], cPath))
 		}
 	}
 
@@ -1234,16 +1250,14 @@ func propagateTransitiveDependents(dependentUnits map[string][]string) {
 
 	maxIterations := len(nodes)
 
-	for i := 0; i < maxIterations; i++ {
+	for range maxIterations {
 		updated := false
 
 		for unit, dependents := range dependentUnits {
 			for _, dep := range dependents {
 				old := dependentUnits[unit]
-				newList := util.RemoveDuplicatesFromList(
-					append(old, dependentUnits[dep]...),
-				)
-				newList = util.RemoveElementFromList(newList, unit)
+				newList := util.RemoveDuplicates(append(old, dependentUnits[dep]...))
+				newList = slices.DeleteFunc(newList, func(path string) bool { return path == unit })
 
 				if len(newList) != len(old) {
 					dependentUnits[unit] = newList
